from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519

from ..CryptoTypes import PrivateKey, PublicKey, Signature


class KeyPair:
	"""Represents an ED25519 private and public key."""

	def __init__(self, private_key):
		"""Creates a key pair from a private key."""
		self._sk = ed25519.Ed25519PrivateKey.from_private_bytes(private_key.bytes)

	@property
	def public_key(self):
		"""Gets the public key."""
		return PublicKey(self._sk.public_key().public_bytes(serialization.Encoding.Raw, serialization.PublicFormat.Raw))

	@property
	def private_key(self):
		"""Gets the private key."""
		return PrivateKey(self._sk.private_bytes(
			encoding=serialization.Encoding.Raw,
			format=serialization.PrivateFormat.Raw,
			encryption_algorithm=serialization.NoEncryption()))

	def sign(self, message):
		"""Signs a message with the private key."""
		return Signature(self._sk.sign(message))


class Verifier:
	"""Verifies signatures signed by a single key pair."""

	def __init__(self, public_key):
		"""Creates a verifier from a public key."""
		if bytes(32) == public_key.bytes:
			raise ValueError('public key cannot be zero')

		self._pk = ed25519.Ed25519PublicKey.from_public_bytes(public_key.bytes)

	def verify(self, message, signature):
		"""Verifies a message signature."""
		try:
			self._pk.verify(signature.bytes, message)
			return True
		except InvalidSignature:
			return False



from mnemonic import Mnemonic


class CodeWords:
	"""Represents a group of words that encode some binary data."""

	def __init__(self, words):
		self.words = words
		self.separator = ' '

	def __eq__(self, other):
		return isinstance(other, CodeWords) and self.words == other.words

	def __str__(self):
		return self.separator.join(self.words)


class CodeWordsEncoder:
	"""Encodes and decodes values using wordlists."""

	def __init__(self, wordlist='english'):
		self.wordlist = wordlist if not isinstance(wordlist, str) else Mnemonic(wordlist).wordlist
		self.wordlist_size = len(self.wordlist)

		if self.wordlist_size < 2:
			raise ValueError('wordlist must contain at least two words')

	def encode_int(self, value):
		"""Encodes an integer into one or more code words."""
		seed = []
		while value:
			index = value % self.wordlist_size
			seed.append(self.wordlist[index])
			value = value // self.wordlist_size

		return CodeWords(seed if seed else [self.wordlist[0]])

	def decode_int(self, code_words):
		"""Decodes one or more code words into an integer."""
		base = 1
		value = 0
		for word in code_words.words:
			index = self.wordlist.index(word)
			value += index * base
			base *= self.wordlist_size

		return value



from .Ordered import Ordered


class BaseValue(Ordered):
	"""Represents a base int."""

	def __init__(self, size, value, tag=None, signed=False):
		"""Creates a base value."""
		self.size = size
		self.value = value
		self.__tag = (tag, signed)

		# check bounds
		bit_size = self.size * 8
		if signed:
			upper_bound = (1 << (bit_size - 1)) - 1
			lower_bound = -upper_bound - 1
		else:
			upper_bound = (1 << bit_size) - 1
			lower_bound = 0

		if self.value < lower_bound or self.value > upper_bound:
			signed_description = 'signed' if signed else 'unsigned'
			value_range_message = f'{value} must be in range [{lower_bound}, {upper_bound}]'
			raise ValueError(f'{value_range_message} for {self.size} bytes ({signed_description})')

	def _cmp(self, other, operation):
		if not isinstance(other, BaseValue):
			return NotImplemented

		# pylint: disable=protected-access
		return operation(self.value, other.value) and self.__tag == other.__tag

	def __eq__(self, other):
		# pylint: disable=protected-access
		return isinstance(other, BaseValue) and self.value == other.value and self.__tag == other.__tag

	def __ne__(self, other):
		return not self == other

	def __hash__(self):
		return hash((self.value, self.__tag))

	def __str__(self):
		if not self.__tag[1] or self.value >= 0:
			unsigned_value = self.value
		else:
			upper_bound_plus_one = 1 << (self.size * 8)
			unsigned_value = self.value + upper_bound_plus_one

		return f'0x{unsigned_value:0{self.size * 2}X}'



import base64
import datetime

import sha3

from ..ByteArray import ByteArray
from ..Network import Network as BasicNetwork
from ..NetworkTimestamp import NetworkTimestamp as BasicNetworkTimestamp
from ..NetworkTimestamp import NetworkTimestampDatetimeConverter


class NetworkTimestamp(BasicNetworkTimestamp):
	"""Represents a nem network timestamp with second resolution."""

	def add_seconds(self, count):
		return NetworkTimestamp(self.timestamp + count)


class Address(ByteArray):
	"""Represents a nem address."""

	SIZE = 25
	ENCODED_SIZE = 40

	def __init__(self, address):
		"""Creates an address from a decoded or encoded address."""
		raw_bytes = address
		if isinstance(address, str):
			raw_bytes = base64.b32decode(address)
		elif isinstance(address, Address):
			raw_bytes = address.bytes

		super().__init__(self.SIZE, raw_bytes, Address)

	def __str__(self):
		return base64.b32encode(self.bytes).decode('utf8')

	def __repr__(self):
		return f'Address(\'{str(self)}\')'


class Network(BasicNetwork):
	"""Represents a nem network."""

	def __init__(self, name, identifier, epoch_time):
		"""Creates a new network with the specified properties."""
		super().__init__(name, identifier, NetworkTimestampDatetimeConverter(epoch_time, 'seconds'), Address, NetworkTimestamp)

	def address_hasher(self):
		return sha3.keccak_256()

	def create_address(self, address_without_checksum, checksum):
		return Address(address_without_checksum + checksum)


Network.MAINNET = Network('mainnet', 0x68, datetime.datetime(2015, 3, 29, 0, 6, 25, tzinfo=datetime.timezone.utc))
Network.TESTNET = Network('testnet', 0x98, datetime.datetime(2015, 3, 29, 0, 6, 25, tzinfo=datetime.timezone.utc))
Network.NETWORKS = [Network.MAINNET, Network.TESTNET]



from binascii import hexlify, unhexlify

from .Ordered import Ordered


class ByteArray(Ordered):
	"""Represents a fixed size byte array."""

	def __init__(self, fixed_size, array_input, tag=None):
		"""Creates a byte array from bytes or hex string."""
		raw_bytes = array_input
		if isinstance(raw_bytes, str):
			raw_bytes = unhexlify(raw_bytes)

		if fixed_size != len(raw_bytes):
			raise ValueError(f'bytes was size {len(raw_bytes)} but must be {fixed_size}')

		self.bytes = raw_bytes
		self.__tag = tag

	def _cmp(self, other, operation):
		if not isinstance(other, ByteArray):
			return NotImplemented

		# pylint: disable=protected-access
		return operation(self.bytes, other.bytes) and self.__tag == other.__tag

	def __eq__(self, other):
		# pylint: disable=protected-access
		return isinstance(other, ByteArray) and self.bytes == other.bytes and self.__tag == other.__tag

	def __ne__(self, other):
		return not self == other

	def __hash__(self):
		return hash(self.bytes)

	def __str__(self):
		return hexlify(self.bytes).decode('utf8').upper()



def read_array_impl(view, factory_class, accessor, should_continue):
	elements = []
	previous_element = None

	i = 0
	while should_continue(i, view):
		element = factory_class.deserialize(view)

		if element.size <= 0:
			raise ValueError('element size has invalid size')

		if accessor and previous_element and accessor(previous_element) >= accessor(element):
			raise ValueError('elements in array are not sorted')

		elements.append(element)
		view = view[element.size:]

		previous_element = element
		i += 1

	return elements


def write_array_impl(elements, count, accessor):
	output_buffer = bytes()
	for i in range(0, count):
		element = elements[i]
		if accessor and i > 0 and accessor(elements[i - 1]) >= accessor(element):
			raise ValueError('array passed to write array is not sorted')

		output_buffer += element.serialize()

	return output_buffer


class ArrayHelpers:
	@staticmethod
	def get_bytes(view, size):
		"""Returns first size bytes of view."""
		if size > len(view):
			raise ValueError(f'size should not exceed {len(view)}. The value of size was: {size}.')

		return view[:size].tobytes()

	@staticmethod
	def align_up(size, alignment):
		"""Calculates aligned size."""
		return (size + alignment - 1) // alignment * alignment

	@staticmethod
	def size(elements, alignment=0, skip_last_element_padding=False):
		"""Calculates size of variable size objects."""
		if not alignment:
			return sum(map(lambda e: e.size, elements))

		if not skip_last_element_padding:
			return sum(map(lambda e: ArrayHelpers.align_up(e.size, alignment), elements))

		return sum(map(lambda e: ArrayHelpers.align_up(e.size, alignment), elements[:-1])) + sum(map(lambda e: e.size, elements[-1:]))

	@staticmethod
	def read_array(view, factory_class, accessor=None):
		"""Reads array of objects."""
		return read_array_impl(view, factory_class, accessor, lambda _, view: len(view) > 0)

	@staticmethod
	def read_array_count(view, factory_class, count, accessor=None):
		"""Reads array of deterministic number of objects."""
		return read_array_impl(view, factory_class, accessor, lambda index, _: count > index)

	@staticmethod
	def read_variable_size_elements(view, factory_class, alignment, skip_last_element_padding=False):
		"""Reads array of variable size objects."""
		elements = []
		while len(view) > 0:
			element = factory_class.deserialize(view)

			if element.size <= 0:
				raise ValueError('element size has invalid size')

			elements.append(element)

			aligned_size = ArrayHelpers.align_up(element.size, alignment)
			if skip_last_element_padding and element.size >= len(view):
				aligned_size = element.size

			if aligned_size > len(view):
				raise ValueError('unexpected buffer length')

			view = view[aligned_size:]

		return elements

	@staticmethod
	def write_array(elements, accessor=None):
		"""Writes array of objects."""
		return write_array_impl(elements, len(elements), accessor)

	@staticmethod
	def write_array_count(elements, count, accessor=None):
		"""Writes array of deterministic number of objects."""
		return write_array_impl(elements, count, accessor)

	@staticmethod
	def write_variable_size_elements(elements, alignment, skip_last_element_padding=False):
		"""Writes array of variable size objects."""
		output_buffer = bytes()
		for index, element in enumerate(elements):
			output_buffer += element.serialize()

			if not skip_last_element_padding or len(elements) - 1 != index:
				aligned_size = ArrayHelpers.align_up(element.size, alignment)
				if aligned_size != element.size:
					output_buffer += bytes(aligned_size - element.size)

		return output_buffer



from .BufferWriter import BufferWriter
from .CryptoTypes import Hash256, Signature
from .QrStorage import QrStorage


class QrSignatureStorage:
	"""Loads and saves signatures as QR codes in a directory."""

	def __init__(self, directory):
		"""Creates storage for a directory."""
		self.storage = QrStorage(directory)

	def save(self, name, transaction_hash, signatures):
		"""Saves a transaction hash along with attesting signatures."""
		writer = BufferWriter()
		writer.write_bytes(transaction_hash.bytes)
		for signature in signatures:
			writer.write_bytes(signature.bytes)

		self.storage.save(name, writer.buffer)

	def load(self, name):
		"""Loads a transaction hash along with attesting signatures."""
		decoded_buffer = self.storage.load(name)

		decoded_buffer_size = len(decoded_buffer)
		if decoded_buffer_size < Hash256.SIZE or 0 != (decoded_buffer_size - Hash256.SIZE) % Signature.SIZE:
			raise ValueError(f'decoded buffer from QR code has unexpected size {decoded_buffer_size}')

		transaction_hash = Hash256(decoded_buffer[0:Hash256.SIZE])

		signatures = []
		for i in range(0, (decoded_buffer_size - Hash256.SIZE) // Signature.SIZE):
			signature_start = Hash256.SIZE + i * Signature.SIZE
			signatures.append(Signature(decoded_buffer[signature_start:signature_start + Signature.SIZE]))

		return (transaction_hash, signatures)



import yaml

from .CryptoTypes import PublicKey


class AccountDescriptor:
	"""Represents an account."""

	def __init__(self, descriptor_yaml):
		"""Creates a descriptor from a yaml container."""
		self.public_key = descriptor_yaml.get('public_key')
		if self.public_key:
			self.public_key = PublicKey(self.public_key)

		self.address = descriptor_yaml.get('address')
		self.name = descriptor_yaml.get('name')
		self.roles = descriptor_yaml.get('roles') or []


class AccountDescriptorRepository:
	"""Loads read-only account descriptors from YAML."""

	def __init__(self, yaml_input):
		"""Loads account descriptors from the specified input."""
		descriptors_yaml = yaml_input if isinstance(yaml_input, list) else yaml.load(yaml_input, Loader=yaml.SafeLoader)
		self.descriptors = [AccountDescriptor(descriptor_yaml) for descriptor_yaml in descriptors_yaml]

	def try_find_by_name(self, name):
		"""Finds the account descriptor with a matching name or None if no matching descriptors are found."""
		return next((descriptor for descriptor in self.descriptors if name == descriptor.name), None)

	def find_by_public_key(self, public_key):
		"""Finds the account descriptor with a matching public key."""
		return next(descriptor for descriptor in self.descriptors if descriptor.public_key and public_key == descriptor.public_key)

	def find_by_address(self, address):
		"""Finds the account descriptor with a matching address."""
		return next(descriptor for descriptor in self.descriptors if descriptor.address and str(address) == descriptor.address)

	def find_all_by_role(self, role):
		"""Finds all account descriptors with a matching role."""
		return [descriptor for descriptor in self.descriptors if not role or role in descriptor.roles]

	def _lookup_account_descriptor_field(self, value, property_name, target_class):
		account_descriptor = self.try_find_by_name(value)
		if account_descriptor and hasattr(account_descriptor, property_name):
			return target_class(getattr(account_descriptor, property_name))

		return target_class(value)

	def _bind_lookup_account_descriptor_field(self, property_name, target_class):
		return lambda value: self._lookup_account_descriptor_field(value, property_name, target_class)

	def to_type_parsing_rules_map(self, type_to_property_mapping):
		"""Builds a type to parsing rule map."""
		return {
			target_class: self._bind_lookup_account_descriptor_field(property_name, target_class)
			for target_class, property_name in type_to_property_mapping.items()
		}



import sha3

NAMESPACE_FLAG = 1 << 63


def generate_mosaic_id(owner_address, nonce):
	"""Generates a mosaic id from an owner address and a nonce."""
	hasher = sha3.sha3_256()
	hasher.update(nonce.to_bytes(4, 'little'))
	hasher.update(owner_address.bytes)
	digest = hasher.digest()

	result = int.from_bytes(digest[0:8], 'little')
	if result & NAMESPACE_FLAG:
		result -= NAMESPACE_FLAG

	return result


def generate_namespace_id(name, parent_namespace_id=0):
	"""Generates a namespace id from a name and an optional parent namespace id."""
	hasher = sha3.sha3_256()
	hasher.update(parent_namespace_id.to_bytes(8, 'little'))
	hasher.update(name.encode('utf8'))
	digest = hasher.digest()

	result = int.from_bytes(digest[0:8], 'little')
	return result | NAMESPACE_FLAG


def generate_mosaic_alias_id(fully_qualified_name):
	"""Generates a mosaic id from a fully qualified mosaic alias name."""
	return generate_namespace_path(fully_qualified_name)[-1]


def is_valid_namespace_name(name):
	"""Returns true if a name is a valid namespace name."""
	def is_alphanum(character):
		return 'a' <= character <= 'z' or '0' <= character <= '9'

	return name and is_alphanum(name[0]) and all(is_alphanum(ch) or ch in ['_', '-'] for ch in name)


def generate_namespace_path(fully_qualified_name):
	"""Parses a fully qualified namespace name into a path."""
	path = []
	parent_namespace_id = 0
	for name in fully_qualified_name.split('.'):
		if not is_valid_namespace_name(name):
			raise ValueError(f'fully qualified name is invalid due to invalid part name ({fully_qualified_name})')

		path.append(generate_namespace_id(name, parent_namespace_id))
		parent_namespace_id = path[-1]

	return path



import base64
import os

import qrcode
from PIL import Image
from pyzbar import pyzbar


class QrStorage:
	"""Loads and saves binary data as QR codes in a directory."""

	def __init__(self, directory):
		"""Creates storage for a directory."""
		self.directory = directory

	def save(self, name, buffer):
		"""Saves a buffer."""
		qrcode_payload = self._binary_to_qrcode_payload(buffer)
		qrcode_builder = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_H)
		qrcode_builder.add_data(qrcode.util.QRData(qrcode_payload, qrcode.util.MODE_ALPHA_NUM))
		qrcode_builder.make(fit=True)

		qrcode_image = qrcode_builder.make_image(fill_color='#44004E', back_color='#FFFFFF')
		qrcode_image.save(self._get_file_path(name))

	def load(self, name):
		"""Loads a buffer."""
		# [0] because exactly one qrcode is expected in the image
		qrcode_payload = pyzbar.decode(Image.open(self._get_file_path(name)), symbols=[pyzbar.ZBarSymbol.QRCODE])[0].data
		return self._qrcode_payload_to_binary(qrcode_payload)

	@staticmethod
	def _binary_to_qrcode_payload(buffer):
		# replace '=' with a character in QR code alphanumeric alphabet, which enables more efficient data packing in the qrcode
		return base64.b32encode(buffer).decode('utf8').replace('=', '$').encode('utf8')

	@staticmethod
	def _qrcode_payload_to_binary(qrcode_payload):
		return base64.b32decode(qrcode_payload.decode('utf8').replace('$', '='))

	def _get_file_path(self, name):
		return os.path.join(self.directory, f'{name}.png')



from binascii import hexlify
from collections import namedtuple
from enum import Enum
from functools import reduce

import sha3

from ..BufferReader import BufferReader
from ..CryptoTypes import Hash256

MerklePart = namedtuple('MerklePart', ['hash', 'is_left'])
PatriciaTreePath = namedtuple('PatriciaTreePath', ['path', 'size'])

# region MerkleHashBuilder


class MerkleHashBuilder:
	"""Builder for creating a merkle hash."""

	def __init__(self):
		"""Creates a merkle hash builder."""
		self.hashes = []

	def update(self, component_hash):
		"""Adds a hash to the merkle hash."""
		self.hashes.append(component_hash.bytes)

	def final(self):
		"""Calculates the merkle hash."""
		if not self.hashes:
			return Hash256.zero()

		num_remaining_hashes = len(self.hashes)
		while num_remaining_hashes > 1:
			i = 0
			while i < num_remaining_hashes:
				hasher = sha3.sha3_256()
				hasher.update(self.hashes[i])

				if i + 1 < num_remaining_hashes:
					hasher.update(self.hashes[i + 1])
				else:
					# if there is an odd number of hashes, duplicate the last one
					hasher.update(self.hashes[i])
					num_remaining_hashes += 1

				self.hashes[i // 2] = hasher.digest()
				i += 2

			num_remaining_hashes //= 2

		return Hash256(self.hashes[0])

# endregion


# region prove_merkle

def prove_merkle(leaf_hash, merkle_path, root_hash):
	"""
	Proves a merkle hash.
	Merkle *hash chain* path is ordered from leaf to root, where each element is MerklePart.
	"""

	def calculate_next_hash(working_hash, merkle_part):
		hasher = sha3.sha3_256()
		if merkle_part.is_left:
			hasher.update(merkle_part.hash.bytes)
			hasher.update(working_hash.bytes)
		else:
			hasher.update(working_hash.bytes)
			hasher.update(merkle_part.hash.bytes)

		return Hash256(hasher.digest())

	computed_root_hash = reduce(calculate_next_hash, merkle_path, leaf_hash)
	return root_hash == computed_root_hash

# endregion


# region LeafNode / BranchNode

def _get_nibble_at(path, index):
	byte = path.path[index // 2]
	return byte & 0xF if 1 == index % 2 else byte >> 4


def _encode_path(path, is_leaf):
	i = 0
	buffer = [0x20 if is_leaf else 0]
	if 1 == path.size % 2:
		buffer[0] |= 0x10 | _get_nibble_at(path, 0)
		i += 1

	while i < path.size:
		buffer.append((_get_nibble_at(path, i) << 4) + _get_nibble_at(path, i + 1))
		i += 2

	return bytes(buffer)


class LeafNode:
	"""Leaf node in a compact patricia tree."""

	def __init__(self, path, value):
		self.path = path
		self.value = value

	def calculate_hash(self):
		"""Calculates node hash."""

		hasher = sha3.sha3_256()
		hasher.update(_encode_path(self.path, True))
		hasher.update(self.value.bytes)
		return Hash256(hasher.digest())


class BranchNode:
	"""Branch node in a compact patricia tree."""

	def __init__(self, path, links):
		self.path = path
		self.links = links

	def calculate_hash(self):
		"""Calculates node hash."""

		hasher = sha3.sha3_256()
		hasher.update(_encode_path(self.path, False))
		for link in self.links:
			hasher.update((link if link else Hash256.zero()).bytes)

		return Hash256(hasher.digest())

# endregion


# region deserialize_patricia_tree_nodes

def _deserialize_path(reader):
	num_nibbles = reader.read_int(1)
	num_bytes = (num_nibbles + 1) // 2
	return PatriciaTreePath(reader.read_bytes(num_bytes), num_nibbles)


def _deserialize_leaf(reader):
	path = _deserialize_path(reader)
	value = Hash256(reader.read_bytes(Hash256.SIZE))
	return LeafNode(path, value)


def _deserialize_branch(reader):
	path = _deserialize_path(reader)

	links_mask = reader.read_int(2)
	links = [None] * 16
	for index, _ in enumerate(links):
		if links_mask & (2 ** index):
			links[index] = Hash256(reader.read_bytes(Hash256.SIZE))

	return BranchNode(path, links)


def deserialize_patricia_tree_nodes(buffer):
	reader = BufferReader(buffer)
	nodes = []
	while not reader.eof:
		node_marker = reader.read_int(1)

		if 0xFF == node_marker:
			nodes.append(_deserialize_leaf(reader))
		elif 0x00 == node_marker:
			nodes.append(_deserialize_branch(reader))
		else:
			raise ValueError(f'invalid marker of a serialized node ({node_marker})')

	return nodes

# endregion


# region prove_patricia_merkle

class PatriciaMerkleProofResult(Enum):
	"""Possible results of a patricia merkle proof."""

	VALID_POSITIVE = 0x0001  # proof is valid (positive)
	VALID_NEGATIVE = 0x0002  # proof is valid (negative)

	INCONCLUSIVE = 0x4001  # negative proof is inconclusive

	STATE_HASH_DOES_NOT_MATCH_ROOTS = 0x8001  # state hash cannot be derived from subcache merkle roots
	UNANCHORED_PATH_TREE = 0x8002  # root of the path tree being proven is not a subcache merkle root
	LEAF_VALUE_MISMATCH = 0x8003  # leaf value does not match expected value
	UNLINKED_NODE = 0x8004  # provided merkle hash contains an unlinked node
	PATH_MISMATCH = 0x8005  # actual merkle path does not match encoded key


def _check_state_hash(state_hash, subcache_merkle_roots):
	hasher = sha3.sha3_256()
	for root in subcache_merkle_roots:
		hasher.update(root.bytes)

	return state_hash == Hash256(hasher.digest())


def prove_patricia_merkle(encoded_key, value_to_test, merkle_path, state_hash, subcache_merkle_roots):
	"""
	Proves a patricia merkle hash.
	Merkle *node* path is ordered from root to leaf, where each element is either BranchNode or LeafNode.
	"""

	# pylint: disable=too-many-return-statements

	if not _check_state_hash(state_hash, subcache_merkle_roots):
		return PatriciaMerkleProofResult.STATE_HASH_DOES_NOT_MATCH_ROOTS

	if merkle_path[0].calculate_hash() not in subcache_merkle_roots:
		return PatriciaMerkleProofResult.UNANCHORED_PATH_TREE

	# positive proof must end with a leaf
	is_positive_proof = hasattr(merkle_path[-1], 'value')
	if is_positive_proof:
		if value_to_test != merkle_path[-1].value:
			return PatriciaMerkleProofResult.LEAF_VALUE_MISMATCH

	child_hash = None
	actual_path = ''
	for node in reversed(merkle_path):
		node_hash = node.calculate_hash()
		formatted_link_index = ''
		if child_hash:
			if child_hash not in node.links:
				return PatriciaMerkleProofResult.UNLINKED_NODE

			formatted_link_index = f'{node.links.index(child_hash):01X}'

		child_hash = node_hash
		actual_path = f'{formatted_link_index}{hexlify(node.path.path).decode("utf8").upper()}{actual_path}'

	if is_positive_proof:
		# for positive proof, expected and calculated paths must match exactly
		return PatriciaMerkleProofResult.PATH_MISMATCH if actual_path != str(encoded_key) else PatriciaMerkleProofResult.VALID_POSITIVE

	# for negative proof, expected path must start with calculated path and next nibble must be a dead end
	if not str(encoded_key).startswith(actual_path):
		return PatriciaMerkleProofResult.PATH_MISMATCH

	next_nibble = _get_nibble_at(PatriciaTreePath(encoded_key.bytes, 2 * len(encoded_key.bytes)), len(actual_path))
	next_node = merkle_path[-1].links[next_nibble]
	return PatriciaMerkleProofResult.INCONCLUSIVE if next_node is not None else PatriciaMerkleProofResult.VALID_NEGATIVE

# endregion



import yaml


class NodeDescriptor:
	"""Represents a node."""

	def __init__(self, descriptor_yaml):
		"""Creates a descriptor from a yaml container."""
		self.host = descriptor_yaml.get('host')
		self.roles = descriptor_yaml.get('roles') or []


class NodeDescriptorRepository:
	"""Loads read-only node descriptors from YAML."""

	def __init__(self, yaml_input):
		"""Loads node descriptors from the specified input."""
		descriptors_yaml = yaml_input if isinstance(yaml_input, list) else yaml.load(yaml_input, Loader=yaml.SafeLoader)
		self.descriptors = [NodeDescriptor(descriptor_yaml) for descriptor_yaml in descriptors_yaml]

	def find_all_by_role(self, role):
		"""Finds all node descriptors with a matching role."""
		return [descriptor for descriptor in self.descriptors if not role or role in descriptor.roles]

	def find_all_not_by_role(self, role):
		"""Finds all node descriptors without a matching role."""
		return [descriptor for descriptor in self.descriptors if role and role not in descriptor.roles]



import datetime
from abc import abstractmethod


class NetworkTimestamp:
	"""Represents a network timestamp."""

	def __init__(self, timestamp):
		"""Creates a timestamp."""
		self.timestamp = timestamp

	@property
	def is_epochal(self):
		"""Determines if this is the epochal timestamp."""
		return 0 == self.timestamp

	@abstractmethod
	def add_seconds(self, count):
		"""Adds a specified number of seconds to this timestamp."""

	def add_minutes(self, count):
		"""Adds a specified number of minutes to this timestamp."""
		return self.add_seconds(60 * count)

	def add_hours(self, count):
		"""Adds a specified number of hours to this timestamp."""
		return self.add_minutes(60 * count)

	def __eq__(self, other):
		return isinstance(other, NetworkTimestamp) and self.timestamp == other.timestamp

	def __str__(self):
		return str(self.timestamp)


class NetworkTimestampDatetimeConverter:
	"""Provides utilities for converting between network timestamps and datetimes."""

	def __init__(self, epoch, time_units):
		"""Creates a converter given an epoch and base time units."""
		self.epoch = epoch
		self.time_units = time_units

	def to_datetime(self, raw_timestamp):
		"""Converts a network timestamp to a datetime."""
		return self.epoch + datetime.timedelta(**{self.time_units: raw_timestamp})

	def to_difference(self, reference_datetime):
		"""Subtracts the network epoch from the reference date."""
		if reference_datetime < self.epoch:
			raise ValueError('timestamp cannot be before epoch')

		return (reference_datetime - self.epoch) / datetime.timedelta(**{self.time_units: 1})



import base64
import datetime

import sha3

from ..ByteArray import ByteArray
from ..CryptoTypes import Hash256
from ..Network import Network as BasicNetwork
from ..NetworkTimestamp import NetworkTimestamp as BasicNetworkTimestamp
from ..NetworkTimestamp import NetworkTimestampDatetimeConverter


class NetworkTimestamp(BasicNetworkTimestamp):
	"""Represents a symbol network timestamp with millisecond resolution."""

	def add_milliseconds(self, count):
		"""Adds a specified number of milliseconds to this timestamp."""
		return NetworkTimestamp(self.timestamp + count)

	def add_seconds(self, count):
		return self.add_milliseconds(1000 * count)


class Address(ByteArray):
	"""Represents a Symbol address."""

	SIZE = 24
	ENCODED_SIZE = 39

	def __init__(self, address):
		"""Creates an address from a decoded or encoded address."""
		raw_bytes = address
		if isinstance(address, str):
			raw_bytes = base64.b32decode(address + 'A')[0:-1]
		elif isinstance(address, Address):
			raw_bytes = address.bytes

		super().__init__(self.SIZE, raw_bytes, Address)

	def __str__(self):
		return base64.b32encode(self.bytes + bytes(0)).decode('utf8')[0:-1]

	def __repr__(self):
		return f'Address(\'{str(self)}\')'


class Network(BasicNetwork):
	"""Represents a Symbol network."""

	def __init__(self, name, identifier, epoch_time, generation_hash_seed=None):
		"""Creates a new network with the specified properties."""
		super().__init__(name, identifier, NetworkTimestampDatetimeConverter(epoch_time, 'milliseconds'), Address, NetworkTimestamp)
		self.generation_hash_seed = generation_hash_seed

	def address_hasher(self):
		return sha3.sha3_256()

	def create_address(self, address_without_checksum, checksum):
		return Address(address_without_checksum + checksum[0:3])


Network.MAINNET = Network(
	'mainnet',
	0x68,
	datetime.datetime(2021, 3, 16, 0, 6, 25, tzinfo=datetime.timezone.utc),
	Hash256('57F7DA205008026C776CB6AED843393F04CD458E0AA2D9F1D5F31A402072B2D6'))
Network.TESTNET = Network(
	'testnet',
	0x98,
	datetime.datetime(2022, 10, 31, 21, 7, 47, tzinfo=datetime.timezone.utc),
	Hash256('49D6E1CE276A85B70EAFE52349AACCA389302E7A9754BCF1221E79494FC665A4'))
Network.MIJINNET = Network(
    'mijinnet',
    0x60,
    datetime.datetime(2019, 6, 11, 23, 0, 0, tzinfo=datetime.timezone.utc),
    Hash256('4BF88474763412AAD36FA13CF7D04894CEE6A54D708053D48466288D5BBF7CA6'))
Network.MIJINTESTNET = Network(
    'mijintestnet',
	0x90,
    datetime.datetime(2019, 6, 11, 23, 0, 0, tzinfo=datetime.timezone.utc),
    Hash256('4BF88474763412AAD36FA13CF7D04894CEE6A54D708053D48466288D5BBF7CA6'))

Network.NETWORKS = [Network.MAINNET, Network.TESTNET, Network.MIJINNET, Network.MIJINTESTNET]



from ..SharedKey import SharedKey as BasicSharedKey


class SharedKey(BasicSharedKey):
	@staticmethod
	def derive_shared_key(key_pair, other_public_key):
		"""Derives shared encryption key from private key and public key."""

		return BasicSharedKey._derive_shared_key(other_public_key.bytes, key_pair.private_key.bytes, b'catapult')



from binascii import unhexlify


class BufferWriter:
	"""Writes data to an in memory buffer."""

	def __init__(self, byte_order='little'):
		"""Creates a writer with specified byte order."""
		self.byte_order = byte_order
		self.buffer = bytes()

	def write_int(self, value, count):
		"""Writes an integer."""
		self.buffer += value.to_bytes(count, self.byte_order)

	def write_string(self, value):
		"""Writes a string."""
		self.buffer += value.encode('utf8')

	def write_hex_string(self, value):
		"""Writes a hex string."""
		self.buffer += unhexlify(value)

	def write_bytes(self, value):
		"""Writes bytes."""
		self.buffer += value



import sha3

from .. import sc
from ..CryptoTypes import Hash256, PublicKey, Signature
from ..Network import NetworkLocator
from ..symbol.KeyPair import KeyPair, Verifier
from ..symbol.Merkle import MerkleHashBuilder
from ..symbol.Network import Address, Network
from ..symbol.SharedKey import SharedKey
from ..symbol.TransactionFactory import TransactionFactory

TRANSACTION_HEADER_SIZE = sum(field[1] for field in [
	('size', 4),
	('reserved1', 4),
	('signature', Signature.SIZE),
	('signer', PublicKey.SIZE),
	('reserved2', 4)
])

AGGREGATE_HASHED_SIZE = sum(field[1] for field in [
	('version_network_type', 4),
	('max_fee', 8),
	('deadline', 8),
	('transactions_hash', Hash256.SIZE)
])


class SymbolFacade:
	"""Facade used to interact with Symbol blockchain."""

	BIP32_CURVE_NAME = 'ed25519'

	Address = Address  # pylint: disable=duplicate-code
	KeyPair = KeyPair  # pylint: disable=duplicate-code
	Verifier = Verifier
	SharedKey = SharedKey

	def __init__(self, network, account_descriptor_repository=None):
		"""Creates a Symbol facade."""
		self.network = NetworkLocator.find_by_name(Network.NETWORKS, network) if isinstance(network, str) else network
		self.account_descriptor_repository = account_descriptor_repository
		self.transaction_factory = self._create_symbol_transaction_factory()

	def _create_symbol_transaction_factory(self):
		type_parsing_rules = None
		if self.account_descriptor_repository:
			type_to_property_mapping = self._create_symbol_type_to_property_mapping()
			type_parsing_rules = self.account_descriptor_repository.to_type_parsing_rules_map(type_to_property_mapping)

		return TransactionFactory(self.network, type_parsing_rules)

	# NOTE: currently `TypeParserBuilder.create_sdk_wrapper`` assumes SDK types are used as keys (although could be `sc` types as well)
	# resulting sdk type will be converted further via `symbol_type_converter`
	@staticmethod
	def _create_symbol_type_to_property_mapping():
		return {
			Address: 'address',
			PublicKey: 'public_key',
		}

	def hash_transaction(self, transaction):
		"""Hashes a Symbol transaction."""
		hasher = sha3.sha3_256()
		hasher.update(transaction.signature.bytes)
		hasher.update(transaction.signer_public_key.bytes)
		hasher.update(self.network.generation_hash_seed.bytes)
		hasher.update(self._transaction_data_buffer(transaction.serialize()))
		return Hash256(hasher.digest())

	def sign_transaction(self, key_pair, transaction):
		"""Signs a Symbol transaction."""
		sign_buffer = self.network.generation_hash_seed.bytes
		sign_buffer += self._transaction_data_buffer(transaction.serialize())
		return key_pair.sign(sign_buffer)

	def verify_transaction(self, transaction, signature):
		"""Verifies a Symbol transaction."""
		verify_buffer = self.network.generation_hash_seed.bytes
		verify_buffer += self._transaction_data_buffer(transaction.serialize())
		return Verifier(transaction.signer_public_key).verify(verify_buffer, signature)

	def cosign_transaction(self, key_pair, transaction, detached=False):
		"""Cosigns a Symbol transaction."""
		transaction_hash = self.hash_transaction(transaction)

		cosignature = sc.DetachedCosignature() if detached else sc.Cosignature()
		if detached:
			cosignature.parent_hash = sc.Hash256(transaction_hash.bytes)

		cosignature.version = 0
		cosignature.signer_public_key = sc.PublicKey(key_pair.public_key.bytes)
		cosignature.signature = sc.Signature(key_pair.sign(transaction_hash.bytes).bytes)
		return cosignature

	@staticmethod
	def hash_embedded_transactions(embedded_transactions):
		"""Hashes embedded transactions of an aggregate."""
		hash_builder = MerkleHashBuilder()
		for embedded_transaction in embedded_transactions:
			hash_builder.update(Hash256(sha3.sha3_256(embedded_transaction.serialize()).digest()))

		return hash_builder.final()

	def bip32_path(self, account_id):
		"""Creates a network compatible BIP32 path for the specified account."""
		return [44, 4343 if 'mainnet' == self.network.name else 1, account_id, 0, 0]

	@staticmethod
	def bip32_node_to_key_pair(bip32_node):
		"""Derives a Symbol KeyPair from a BIP32 node."""
		return KeyPair(bip32_node.private_key)

	@staticmethod
	def _is_aggregate_transaction(transaction_buffer):
		transaction_type_offset = TRANSACTION_HEADER_SIZE + 2  # skip version and network byte
		transaction_type = (transaction_buffer[transaction_type_offset + 1] << 8) + transaction_buffer[transaction_type_offset]
		aggregate_types = [sc.TransactionType.AGGREGATE_BONDED.value, sc.TransactionType.AGGREGATE_COMPLETE.value]
		return transaction_type in aggregate_types

	@staticmethod
	def _transaction_data_buffer(transaction_buffer):
		data_buffer_start = TRANSACTION_HEADER_SIZE
		data_buffer_end = len(transaction_buffer)
		if SymbolFacade._is_aggregate_transaction(transaction_buffer):
			data_buffer_end = TRANSACTION_HEADER_SIZE + AGGREGATE_HASHED_SIZE

		return transaction_buffer[data_buffer_start:data_buffer_end]



from binascii import hexlify

from symbolchain import nc

from ..CryptoTypes import Hash256, PublicKey
from ..RuleBasedTransactionFactory import RuleBasedTransactionFactory
from .Network import Address


class TransactionFactory:
	"""Factory for creating NEM transactions."""

	def __init__(self, network, type_rule_overrides=None):
		"""Creates a factory for the specified network."""
		self.factory = self._build_rules(type_rule_overrides)
		self.network = network

	def create(self, transaction_descriptor, autosort=True):
		"""
		Creates a transaction from a transaction descriptor.
		When autosort is set (default), descriptor arrays requiring ordering will be automatically sorted.
		When unset, descriptor arrays will be presumed to be already sorted.
		"""
		transaction = self.factory.create_from_factory(nc.TransactionFactory.create_by_name, {
			**transaction_descriptor,
			'network': self.network.identifier
		})
		if autosort:
			transaction.sort()

		# hack: explicitly translate transfer message
		if nc.TransactionType.TRANSFER == transaction.type_ and transaction.message and isinstance(transaction.message.message, str):
			transaction.message.message = transaction.message.message.encode('utf8')

		return transaction

	@staticmethod
	def to_non_verifiable_transaction(transaction):
		"""Converts a transaction to a non-verifiable transaction."""
		non_verifiable_class_name = type(transaction).__name__
		if not non_verifiable_class_name.startswith('NonVerifiable'):
			non_verifiable_class_name = f'NonVerifiable{non_verifiable_class_name}'

		non_verifiable_class = getattr(nc, non_verifiable_class_name)
		non_verifiable_transaction = non_verifiable_class()
		for key in dir(non_verifiable_transaction):
			# isupper() to quickly filter out class properties like TRANSACTION_VERSION or TYPE_HINTS
			if key.startswith('_') or key[0].isupper() or key in ('size', 'serialize', 'deserialize') or key.endswith('_computed'):
				continue

			setattr(non_verifiable_transaction, key, getattr(transaction, key))

		return non_verifiable_transaction

	@staticmethod
	def attach_signature(transaction, signature):
		"""Attaches a signature to a transaction."""
		transaction.signature = nc.Signature(signature.bytes)

		transaction_hex = hexlify(TransactionFactory.to_non_verifiable_transaction(transaction).serialize()).decode('utf8').upper()
		signature_hex = str(signature)
		json_payload = f'{{"data":"{transaction_hex}", "signature":"{signature_hex}"}}'
		return json_payload

	@staticmethod
	def _nem_type_converter(value):
		if isinstance(value, Address):
			# yes, unfortunately, nem's Address is 40 bytes string, but we need to pass it as actual bytes not to confuse ByteArray
			return nc.Address(str(value).encode('utf8'))

		return None

	@staticmethod
	def _build_rules(type_rule_overrides):
		factory = RuleBasedTransactionFactory(nc, TransactionFactory._nem_type_converter, type_rule_overrides)
		factory.autodetect()

		struct_names = [
			'CosignatureV1', 'Message', 'NamespaceId', 'MosaicId', 'Mosaic', 'SizePrefixedMosaic', 'MosaicLevy',
			'MosaicProperty', 'SizePrefixedMosaicProperty', 'MosaicDefinition',
			'MultisigAccountModification', 'SizePrefixedMultisigAccountModification', 'SizePrefixedCosignatureV1'
		]
		for name in struct_names:
			factory.add_struct_parser(name)

		sdk_type_mapping = {
			'Address': Address,
			'Hash256': Hash256,
			'PublicKey': PublicKey,
		}
		for name, typename in sdk_type_mapping.items():
			factory.add_pod_parser(name, typename)

		array_names = [
			'SizePrefixedMosaic', 'SizePrefixedMosaicProperty', 'SizePrefixedMultisigAccountModification', 'SizePrefixedCosignatureV1'
		]
		for name in array_names:
			factory.add_array_parser(f'struct:{name}')

		return factory



class TransactionDescriptorProcessor:
	"""Processes and looks up transaction descriptor properties."""

	def __init__(self, transaction_descriptor, type_parsing_rules, type_converter=None):
		"""Creates a transaction descriptor processor."""
		self.transaction_descriptor = transaction_descriptor
		self.type_parsing_rules = type_parsing_rules or {}
		self.type_converter = type_converter or (lambda x: x)
		self.type_hints = {}

	def _lookup_value_and_apply_type_hints(self, key):
		if key not in self.transaction_descriptor:
			raise ValueError(f'transaction descriptor does not have attribute {key}')

		value = self.transaction_descriptor[key]
		type_hint = self.type_hints.get(key)
		if type_hint in self.type_parsing_rules:
			value = self.type_parsing_rules[type_hint](value)

		return value

	def lookup_value(self, key):
		"""Looks up the value for key."""
		value = self._lookup_value_and_apply_type_hints(key)
		if not any(isinstance(value, type_class) for type_class in [str, bytes, tuple]) and hasattr(value, '__iter__'):
			return [self.type_converter(item) for item in value]

		return self.type_converter(value)

	def copy_to(self, transaction, ignore_keys=None):
		"""Copies all descriptor information to a transaction."""
		for key in self.transaction_descriptor.keys():
			if ignore_keys and key in ignore_keys:
				continue

			if key.endswith('_computed'):
				raise ValueError(f'cannot explicitly set computed field {key}')

			if not hasattr(transaction, key):
				raise ValueError(f'transaction does not have attribute {key}')

			value = self.lookup_value(key)
			if isinstance(value, list):
				getattr(transaction, key).extend(value)
			else:
				setattr(transaction, key, value)

	def set_type_hints(self, type_hints):
		"""Sets type hints."""
		self.type_hints = type_hints or {}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['symbolchain',
   'symbolchain.facade',
   'symbolchain.nc',
   'symbolchain.nem',
   'symbolchain.external',
   'symbolchain.sc',
   'symbolchain.symbol']

package_data = \
{'': ['*']}

install_requires = \
['cryptography==38.0.1',
    'mnemonic==0.20',
    'Pillow==9.2.0',
    'pynacl==1.5.0',
    'pysha3==1.0.2',
    'PyYAML==6.0',
    'pyzbar==0.1.9',
    'ripemd-hash==1.0.0',
    'qrcode==7.3.1']

setup_kwargs = {
    'name': 'techbureau-symbol-sdk-python',
    'version': '3.0.3.dev3',
    'description': 'Symbol SDK',
    'long_description': 'This is symbol project core sdk python library.',
    'author': 'Techbureau Contributors',
    'author_email': 'development@techbureau.jp',
    'maintainer': 'Techbureau Contributors',
    'maintainer_email': 'development@techbureau.jp',
    'url': 'https://github.com/tech-bureau-jp/symbol/tree/dev/sdk/python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}

setup(**setup_kwargs)



from binascii import unhexlify

import cryptography

from symbolchain.CryptoTypes import PrivateKey, PublicKey
from symbolchain.impl.CipherHelpers import decode_aes_gcm, encode_aes_gcm
from symbolchain.symbol.KeyPair import KeyPair
from symbolchain.symbol.SharedKey import SharedKey

DELEGATION_MARKER = unhexlify('FE2A8061577301E2')


class MessageEncoder:
	"""Encrypts and encodes messages between two parties."""

	def __init__(self, key_pair: KeyPair):
		"""Creates message encoder around key pair."""
		self.key_pair = key_pair

	def try_decode(self, recipient_public_key, encoded_message):
		"""Tries to decode encoded message, returns tuple:
		* True, message - if message has been decoded and decrypted
		* False, encoded_message - otherwise
		"""
		if 1 == encoded_message[0]:
			try:
				message = decode_aes_gcm(SharedKey, self.key_pair, recipient_public_key, encoded_message[1:])
				return True, message
			except cryptography.exceptions.InvalidTag:
				pass
		elif 0xFE == encoded_message[0] and DELEGATION_MARKER == encoded_message[:8]:
			try:
				ephemeral_public_key = PublicKey(encoded_message[len(DELEGATION_MARKER):len(DELEGATION_MARKER) + PublicKey.SIZE])
				message = decode_aes_gcm(SharedKey, self.key_pair, ephemeral_public_key, encoded_message[len(DELEGATION_MARKER) + PublicKey.SIZE:])
				return True, message
			except cryptography.exceptions.InvalidTag:
				pass
			except ValueError as exception:
				exceptions = ['point is not in main subgroup']
				if not any(map(lambda message: message in str(exception), exceptions)):
					raise

		return False, encoded_message

	@staticmethod
	def encode_persistent_harvesting_delegation(node_public_key, remote_key_pair, vrf_root_key_pair):  # pylint: disable=invalid-name
		"""Encodes persistent harvesting delegation to node."""
		ephemeral_key_pair = KeyPair(PrivateKey.random())

		message = remote_key_pair.private_key.bytes + vrf_root_key_pair.private_key.bytes
		tag, initialization_vector, cipher_text = encode_aes_gcm(SharedKey, ephemeral_key_pair, node_public_key, message)

		return DELEGATION_MARKER + ephemeral_key_pair.public_key.bytes + tag + initialization_vector + cipher_text

	def encode(self, recipient_public_key: PublicKey, message: bytes):
		"""Encodes message to recipient using recommended format."""

		tag, initialization_vector, cipher_text = encode_aes_gcm(SharedKey, self.key_pair, recipient_public_key, message)
		return b'\1' + tag + initialization_vector + cipher_text



from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

# region AesCbcCipher


class AesCbcCipher:
	"""Performs AES CBC encryption and decryption with a given key."""

	def __init__(self, aes_key):
		"""Creates a cipher around an aes shared key."""

		self.key = aes_key

	def encrypt(self, clear_text, iv):  # pylint: disable=invalid-name
		"""Encrypts clear text."""

		# input needs to be padded in CBC mode to a multiple of block size
		padder = padding.PKCS7(128).padder()
		padded_clear_text = padder.update(clear_text)
		padded_clear_text += padder.finalize()

		encryptor = Cipher(algorithms.AES(self.key.bytes), modes.CBC(iv)).encryptor()
		return encryptor.update(padded_clear_text) + encryptor.finalize()

	def decrypt(self, cipher_text, iv):  # pylint: disable=invalid-name
		"""Decrypts cipher text."""

		decryptor = Cipher(algorithms.AES(self.key.bytes), modes.CBC(iv)).decryptor()
		clear_text = decryptor.update(cipher_text) + decryptor.finalize()

		unpadder = padding.PKCS7(128).unpadder()
		unpadded_cipher_text = unpadder.update(clear_text)
		unpadded_cipher_text += unpadder.finalize()
		return unpadded_cipher_text

# endregion


# region AesGcmCipher

class AesGcmCipher:
	"""Performs AES GCM encryption and decryption with a given key."""

	TAG_SIZE = 16

	def __init__(self, aes_key):
		"""Creates a cipher around an aes shared key."""

		self.key = aes_key

	def encrypt(self, clear_text, iv):  # pylint: disable=invalid-name
		"""Encrypts clear text and appends tag to encrypted payload."""

		encryptor = Cipher(algorithms.AES(self.key.bytes), modes.GCM(iv)).encryptor()
		cipher_text = encryptor.update(clear_text) + encryptor.finalize()
		return cipher_text + encryptor.tag

	def decrypt(self, cipher_text, iv):  # pylint: disable=invalid-name
		"""Decrypts cipher text with appended tag."""

		tag_start_offset = len(cipher_text) - self.TAG_SIZE
		decryptor = Cipher(algorithms.AES(self.key.bytes), modes.GCM(iv, cipher_text[tag_start_offset:])).decryptor()
		return decryptor.update(cipher_text[:tag_start_offset]) + decryptor.finalize()

# endregion



import sha3

from ..CryptoTypes import Hash256, PrivateKey, PublicKey
from ..nem.KeyPair import KeyPair, Verifier
from ..nem.Network import Address, Network
from ..nem.SharedKey import SharedKey
from ..nem.TransactionFactory import TransactionFactory
from ..Network import NetworkLocator


class NemFacade:
	"""Facade used to interact with NEM blockchain."""

	BIP32_CURVE_NAME = 'ed25519-keccak'

	Address = Address  # pylint: disable=duplicate-code
	KeyPair = KeyPair  # pylint: disable=duplicate-code
	Verifier = Verifier
	SharedKey = SharedKey

	def __init__(self, network, account_descriptor_repository=None):
		"""Creates a NEM facade."""
		self.network = NetworkLocator.find_by_name(Network.NETWORKS, network) if isinstance(network, str) else network
		self.account_descriptor_repository = account_descriptor_repository
		self.transaction_factory = self._create_nem_transaction_factory()

	def _create_nem_transaction_factory(self):
		type_parsing_rules = None
		if self.account_descriptor_repository:
			type_parsing_rules = self.account_descriptor_repository.to_type_parsing_rules_map(self._create_nem_type_to_property_mapping())

		return TransactionFactory(self.network, type_parsing_rules)

	# NOTE: currently `TypeParserBuilder.create_sdk_wrapper`` assumes SDK types are used as keys (although could be `nc` types as well)
	# resulting sdk type will be converted further via `nem_type_converter`
	@staticmethod
	def _create_nem_type_to_property_mapping():
		return {
			Address: 'address',
			PublicKey: 'public_key'
		}

	@staticmethod
	def hash_transaction(transaction):
		"""Hashes a NEM transaction."""
		non_verifiable_transaction = TransactionFactory.to_non_verifiable_transaction(transaction)
		return Hash256(sha3.keccak_256(non_verifiable_transaction.serialize()).digest())

	@staticmethod
	def sign_transaction(key_pair, transaction):
		"""Signs a NEM transaction."""
		non_verifiable_transaction = TransactionFactory.to_non_verifiable_transaction(transaction)
		return key_pair.sign(non_verifiable_transaction.serialize())

	@staticmethod
	def verify_transaction(transaction, signature):
		"""Verifies a NEM transaction."""
		non_verifiable_transaction = TransactionFactory.to_non_verifiable_transaction(transaction)
		return Verifier(transaction.signer_public_key).verify(non_verifiable_transaction.serialize(), signature)

	def bip32_path(self, account_id):
		"""Creates a network compatible BIP32 path for the specified account."""
		return [44, 43 if 'mainnet' == self.network.name else 1, account_id, 0, 0]

	@staticmethod
	def bip32_node_to_key_pair(bip32_node):
		"""Derives a NEM KeyPair from a BIP32 node."""
		# BIP32 private keys should be used as is, so reverse here to counteract reverse in KeyPair
		return KeyPair(PrivateKey(bip32_node.private_key.bytes[::-1]))



import os

import yaml

from ..CryptoTypes import PublicKey


class BatchOperations:
	"""Provides facade-based batch operations."""

	class PrepareError(Exception):
		"""Error that is thrown when a batch prepare operation fails."""

		def __init__(self, message):
			"""Creates a new error."""
			Exception.__init__(self, message)

	def __init__(self, facade, output_file_prefix=''):
		"""Creates batch operations around the specified facade."""
		self.facade = facade
		self.output_file_prefix = output_file_prefix

	def load_all(self, transactions_yaml_input):
		"""Loads all transactions from YAML."""
		return [
			self.facade.transaction_factory.create(transaction_descriptor)
			for transaction_descriptor in yaml.load(transactions_yaml_input, Loader=yaml.SafeLoader)
		]

	def _sign_one(self, transaction, private_key_storage, signature_storage, output_filename):
		# transaction.signer_public_key is of different PublicKey type, wrap it in sdk type
		signer_public_key = PublicKey(transaction.signer_public_key.bytes)
		signer_account_name = self.facade.account_descriptor_repository.find_by_public_key(signer_public_key).name
		signer_private_key = private_key_storage.load(signer_account_name)

		signature = self.facade.sign_transaction(self.facade.KeyPair(signer_private_key), transaction)
		transaction_hash = self.facade.hash_transaction(transaction)
		signature_storage.save(output_filename, transaction_hash, [signature])

	def sign_all(self, transactions, private_key_storage, signature_storage):
		"""Signs multiple transactions and saves the signatures as multiple qrcodes."""
		for i, transaction in enumerate(transactions):
			self._sign_one(transaction, private_key_storage, signature_storage, f'sig_{self.output_file_prefix}{i}')

	def prepare_all(self, transactions, signature_storage, output_directory):
		"""Prepares multiple transactions by attaching signatures to each and producing files that can be sent to the network."""
		transaction_count = len(transactions)
		signature_groups = [signature_storage.load(f'sig_{self.output_file_prefix}{i}') for i in range(0, transaction_count)]

		for i, transaction in enumerate(transactions):
			(signed_transaction_hash, signatures) = signature_groups[i]

			transaction_hash = self.facade.hash_transaction(transaction)
			if signed_transaction_hash != transaction_hash:
				raise self.PrepareError(f'transaction hash at {i} does not match signed transaction hash')

			if not self.facade.verify_transaction(transaction, signatures[0]):
				raise self.PrepareError(f'transaction signature at {i} does not verify')

		for i, transaction in enumerate(transactions):
			(_, signatures) = signature_groups[i]

			prepared_transaction_buffer = self.facade.transaction_factory.attach_signature(transaction, signatures[0])
			file_path = os.path.join(output_directory, f'payload_{self.output_file_prefix}{i}.json')
			with open(file_path, 'wb') as outfile:
				outfile.write(prepared_transaction_buffer.encode('utf8'))



def metadata_update_value(old_value, new_value):
	"""Creates a metadata payload for updating old value to new value."""

	if not old_value:
		return new_value

	shorter_length = min(len(old_value), len(new_value))
	longer_length = max(len(old_value), len(new_value))
	is_new_value_shorter = len(old_value) > len(new_value)

	result = bytearray()
	for i in range(shorter_length):
		result.append(old_value[i] ^ new_value[i])

	for i in range(shorter_length, longer_length):
		result.append((old_value if is_new_value_shorter else new_value)[i])

	return bytes(result)



from binascii import hexlify

from symbolchain import sc

from ..CryptoTypes import Hash256, PublicKey
from ..RuleBasedTransactionFactory import RuleBasedTransactionFactory
from .IdGenerator import generate_mosaic_id, generate_namespace_id
from .Network import Address


class TransactionFactory:
	"""Factory for creating Symbol transactions."""

	def __init__(self, network, type_rule_overrides=None):
		"""Creates a factory for the specified network."""
		self.factory = self._build_rules(type_rule_overrides)
		self.network = network

	def _create_and_extend(self, transaction_descriptor, autosort, factory_class):
		transaction = self.factory.create_from_factory(factory_class.create_by_name, {
			**transaction_descriptor,
			'network': self.network.identifier
		})
		if autosort:
			transaction.sort()

		# autogenerate artifact ids
		if sc.TransactionType.NAMESPACE_REGISTRATION == transaction.type_:
			parent_id = transaction.parent_id.value if sc.NamespaceRegistrationType.CHILD == transaction.registration_type else 0
			transaction.id = sc.NamespaceId(generate_namespace_id(transaction.name.decode('utf8'), parent_id))
		elif sc.TransactionType.MOSAIC_DEFINITION == transaction.type_:
			address = self.network.public_key_to_address(PublicKey(transaction.signer_public_key.bytes))
			transaction.id = sc.MosaicId(generate_mosaic_id(address, transaction.nonce.value))

		return transaction

	def create(self, transaction_descriptor, autosort=True):
		"""
		Creates a transaction from a transaction descriptor.
		When autosort is set (default), descriptor arrays requiring ordering will be automatically sorted.
		When unset, descriptor arrays will be presumed to be already sorted.
		"""
		return self._create_and_extend(transaction_descriptor, autosort, sc.TransactionFactory)

	def create_embedded(self, transaction_descriptor, autosort=True):
		"""
		Creates an embedded transaction from a transaction descriptor.
		When autosort is set (default), descriptor arrays requiring ordering will be automatically sorted.
		When unset, descriptor arrays will be presumed to be already sorted.
		"""
		return self._create_and_extend(transaction_descriptor, autosort, sc.EmbeddedTransactionFactory)

	@staticmethod
	def attach_signature(transaction, signature):
		"""Attaches a signature to a transaction."""
		transaction.signature = sc.Signature(signature.bytes)

		transaction_buffer = transaction.serialize()
		hex_payload = hexlify(transaction_buffer).decode('utf8').upper()
		json_payload = f'{{"payload": "{hex_payload}"}}'
		return json_payload

	@staticmethod
	def _symbol_type_converter(value):
		if isinstance(value, Address):
			return sc.UnresolvedAddress(value.bytes)

		return None

	@staticmethod
	def _build_rules(type_rule_overrides):
		factory = RuleBasedTransactionFactory(sc, TransactionFactory._symbol_type_converter, type_rule_overrides)
		factory.autodetect()

		factory.add_struct_parser('UnresolvedMosaic')

		sdk_type_mapping = {
			'UnresolvedAddress': Address,
			'Address': Address,
			'Hash256': Hash256,
			'PublicKey': PublicKey,
			'VotingPublicKey': PublicKey,
		}
		for name, typename in sdk_type_mapping.items():
			factory.add_pod_parser(name, typename)

		for name in ['UnresolvedMosaicId', 'TransactionType', 'UnresolvedAddress', 'struct:UnresolvedMosaic']:
			factory.add_array_parser(name)

		return factory



from abc import abstractmethod

from symbolchain.ripemd160 import ripemd160

BASE32_RFC4648_ALPHABET = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'


class Network:
	"""Represents a network."""

	def __init__(self, name, identifier, datetime_converter, address_class, network_timestamp_class):
		# pylint: disable=too-many-arguments

		"""Creates a new network with the specified properties."""
		self.name = name
		self.identifier = identifier
		self.datetime_converter = datetime_converter

		self.address_class = address_class
		self.network_timestamp_class = network_timestamp_class

	def public_key_to_address(self, public_key):
		"""Converts a public key to an address."""
		part_one_hash_builder = self.address_hasher()
		part_one_hash_builder.update(public_key.bytes)
		part_one_hash = part_one_hash_builder.digest()

		part_two_hash = ripemd160(part_one_hash)

		version = bytes([self.identifier]) + part_two_hash

		part_three_hash_builder = self.address_hasher()
		part_three_hash_builder.update(version)
		checksum = part_three_hash_builder.digest()[0:4]

		return self.create_address(version, checksum)

	def is_valid_address_string(self, address_string):
		"""Checks if an address string is valid and belongs to this network."""
		if self.address_class.ENCODED_SIZE != len(address_string):
			return False

		if any(ch not in BASE32_RFC4648_ALPHABET for ch in address_string):
			return False

		return self.is_valid_address(self.address_class(address_string))

	def is_valid_address(self, address):
		"""Checks if an address is valid and belongs to this network."""
		if address.bytes[0] != self.identifier:
			return False

		hash_builder = self.address_hasher()
		hash_builder.update(address.bytes[0:1 + 20])

		checksum_from_address = address.bytes[1 + 20:]
		calculated_checksum = hash_builder.digest()[0:len(checksum_from_address)]
		return checksum_from_address == calculated_checksum

	def to_datetime(self, reference_network_timestamp):
		"""Converts a network timestamp to a datetime."""
		return self.datetime_converter.to_datetime(reference_network_timestamp.timestamp)

	def from_datetime(self, reference_datetime):
		"""Converts a datetime to a network timestamp."""
		return self.network_timestamp_class(self.datetime_converter.to_difference(reference_datetime))

	@abstractmethod
	def address_hasher(self):
		"""Gets the primary hasher to use in the public key to address conversion."""

	@abstractmethod
	def create_address(self, address_without_checksum, checksum):
		"""Creates an encoded address from an address without checksum and checksum bytes."""

	def __eq__(self, other):
		return isinstance(other, Network) and self.name == other.name and self.identifier == other.identifier

	def __str__(self):
		return self.name


class NetworkLocator:
	"""Provides utility functions for finding a network."""

	@staticmethod
	def find_by_name(networks, names):
		"""Finds a network with a specified name within a list of networks."""
		if isinstance(names, str):
			names = [names]

		return next(network for network in networks if network.name in names)

	@staticmethod
	def find_by_identifier(networks, identifiers):
		"""Finds a network with a specified identifier within a list of networks."""
		if isinstance(identifiers, int):
			identifiers = [identifiers]

		return next(network for network in networks if network.identifier in identifiers)



from ..BufferWriter import BufferWriter
from ..CryptoTypes import PrivateKey
from .KeyPair import KeyPair


class VotingKeysGenerator:
	"""Generates symbol voting keys."""

	def __init__(self, root_key_pair, private_key_generator=PrivateKey.random):
		"""Creates a generator around a voting root key pair."""
		self.root_key_pair = root_key_pair
		self.private_key_generator = private_key_generator

	def generate(self, start_epoch, end_epoch):
		"""Generates voting keys for specified epochs."""
		writer = BufferWriter()
		writer.write_int(start_epoch, 8)  # start key identifier
		writer.write_int(end_epoch, 8)  # end key identifier
		writer.write_int(0xFFFFFFFFFFFFFFFF, 8)  # reserved - last (used) key identifier
		writer.write_int(0xFFFFFFFFFFFFFFFF, 8)  # reserved - last wiped key identifier

		writer.write_bytes(self.root_key_pair.public_key.bytes)  # root voting public key
		writer.write_int(start_epoch, 8)  # level 1/1 start key identifier
		writer.write_int(end_epoch, 8)  # level 1/1 end key identifier

		for identifier in reversed(range(start_epoch, end_epoch + 1)):
			child_private_key = self.private_key_generator()
			child_key_pair = KeyPair(child_private_key)

			parent_signed_payload_writer = BufferWriter()
			parent_signed_payload_writer.write_bytes(child_key_pair.public_key.bytes)
			parent_signed_payload_writer.write_int(identifier, 8)
			signature = self.root_key_pair.sign(parent_signed_payload_writer.buffer)

			writer.write_bytes(child_key_pair.private_key.bytes)  # child voting private key used to sign votes for an epoch
			writer.write_bytes(signature.bytes)  # signature proving derivation of child key pair from root

		return writer.buffer



import os

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519

from .CryptoTypes import PrivateKey


class PrivateKeyStorage:
	"""Loads and saves private keys as PEM files with optional encryption in a directory."""

	def __init__(self, directory, password=None):
		"""Creates storage for a directory."""
		self.directory = directory
		self.password = None if not password else password.encode('utf8')

	def save(self, name, private_key):
		"""Saves a private key with the specified name."""
		wrapped_private_key = ed25519.Ed25519PrivateKey.from_private_bytes(private_key.bytes)
		with open(self._get_file_path(name), 'wb') as outfile:
			encryption_algorithm = serialization.NoEncryption()
			if self.password:
				encryption_algorithm = serialization.BestAvailableEncryption(self.password)

			outfile.write(wrapped_private_key.private_bytes(
				encoding=serialization.Encoding.PEM,
				format=serialization.PrivateFormat.PKCS8,
				encryption_algorithm=encryption_algorithm))

	def load(self, name):
		"""Loads a private key with the specified name."""
		with open(self._get_file_path(name), 'rb') as infile:
			wrapped_private_key = serialization.load_pem_private_key(infile.read(), password=self.password)
			private_key_bytes = wrapped_private_key.private_bytes(
				encoding=serialization.Encoding.Raw,
				format=serialization.PrivateFormat.Raw,
				encryption_algorithm=serialization.NoEncryption())
			return PrivateKey(private_key_bytes)

	def _get_file_path(self, name):
		return os.path.join(self.directory, f'{name}.pem')



import sha3

from ..CryptoTypes import SharedKey256
from ..external import ed25519
from ..SharedKey import SharedKey as BasicSharedKey


class SharedKey(BasicSharedKey):
	@staticmethod
	def derive_shared_key(key_pair, other_public_key):
		"""Derives shared encryption key from private key and public key."""

		# note: key_pair.private_key is "unreversed", so need to reverse here again
		return BasicSharedKey._derive_shared_key(other_public_key.bytes, key_pair.private_key.bytes[::-1], b'nem-nis1', sha3.keccak_512)

	@staticmethod
	def derive_shared_key_deprecated(key_pair, other_public_key, salt):
		"""Derives shared encryption key from private key, public key and salt.

		note: this method uses _old_ method of deriving shared key and has been deprecated.
		"""

		# note: key_pair.private_key is "unreversed", so need to reverse here again
		shared_secret = ed25519.derive_shared_secret_unsafe(other_public_key.bytes, key_pair.private_key.bytes[::-1], sha3.keccak_512)
		key = bytes([shared_secret[i] ^ salt[i] for i in range(32)])
		return SharedKey256(sha3.keccak_256(key).digest())
